#include "cinder/app/AppBasic.h"
#include "cinder/gl/gl.h"
#include "cinder/Utilities.h"
#include "cinder/Rand.h"
#include <list>

#include "CinderOpenCv.h"

#define POINTS_NUMBER 2000

using namespace ci;
using namespace ci::app;
using namespace std;

class DelaunayLineApp : public AppBasic {
  public:
	struct LinePos {
		Vec2f first, second;
		LinePos(Vec2f _first, Vec2f _second)
		: first(_first), second(_second){}
	};

	void prepareSettings( Settings *settings );
	void setup();
	void shutdown();
	void mouseDown( MouseEvent event );	
	void update();
	void draw();
	void initDelaunayStorage();
	list<LinePos> getDelaunayLines();

	
	Color						backGround;
	ColorA					drawColor;
	
	CvMemStorage*		storage;
	CvSubdiv2D*			subdiv;
	
	vector<Vec2f>		points;
	list<LinePos>	lines;
	
};

void DelaunayLineApp::prepareSettings( Settings *settings )
{
	settings->setWindowSize( 640, 480 );
	settings->setFrameRate( 60.0f );
	backGround = Color(1,1,1);
	drawColor = ColorA( 0.0f, 0.0f, 0.0f, 0.1f );
}

void DelaunayLineApp::setup()
{

	// Generate Points
	for(int i = 0; i < POINTS_NUMBER; i++){
		points.push_back(Vec2f::zero());
	}
	
	// GL setup
	gl::enableAlphaBlending();
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_POINT_SMOOTH);
	gl::clear( backGround, true );
	glColor4f( drawColor );
	glLineWidth(0.1);
	
}

void DelaunayLineApp::shutdown()
{
	cvReleaseMemStorage( &storage );
}

void DelaunayLineApp::mouseDown( MouseEvent event )
{
	try {
		writeImage( getHomeDirectory() + "cinder" + getPathSeparator() + "DelaunayLineApp_saveImage.png", copyWindowSurface() );
	}
	catch( ... ) {
		console() << "unable to save the image file!" << std::endl;
	}		
	
}

void DelaunayLineApp::update()
{

	// set points
	for(int i = 0; i < points.size(); i++){
		float posx = 	Rand::randFloat()*Rand::randFloat(-1,1)*getWindowWidth()*0.5+getWindowWidth()*0.5;
		float posy = 	Rand::randFloat(-1,1)*Rand::randFloat(-1,1)*getWindowHeight()*0.5+getWindowHeight()*0.5;
		points[i]=Vec2f(posx,posy);
	}
	
	// calc Delaunay
	initDelaunayStorage();
	CvSubdiv2DPoint *p;
	for(int i = 0; i < points.size(); i++){
		p = cvSubdivDelaunay2DInsert( subdiv, toOcv(points[i]) );
		p->flags = i;
	}
	
	// get lines
	lines.clear();
	lines = getDelaunayLines();
	
	cvReleaseMemStorage( &storage );

}

void DelaunayLineApp::draw()
{
	// clear out the window with backGround
	gl::clear( backGround, true );
	
	//draw points
	for(int i = 0; i < points.size(); i++){
		gl::drawSolidCircle(points[i],2);
	}

	//draw_lines
	for( list<LinePos>::iterator it = lines.begin(); it != lines.end(); ++it ){
		gl::drawLine(it->first, it->second);
	}
	
}

//-------------------------------------------------------------------
// Delaunay
//-------------------------------------------------------------------
void DelaunayLineApp::initDelaunayStorage()
{
	CvRect rect = { 0, 0, getWindowWidth(), getWindowHeight() };
	storage = cvCreateMemStorage(0);
	
	subdiv = cvCreateSubdiv2D( CV_SEQ_KIND_SUBDIV2D, sizeof(*subdiv),
														sizeof(CvSubdiv2DPoint),
														sizeof(CvQuadEdge2D),
														storage );
	cvInitSubdivDelaunay2D( subdiv, rect );
}

list<DelaunayLineApp::LinePos> DelaunayLineApp::getDelaunayLines()
{
	list<LinePos> temp_lines;
	
	int i, org_id, dst_id;
	int total = subdiv->edges->total;
	int elem_size = subdiv->edges->elem_size;
	
	CvSubdiv2DPoint* org_pt;
	CvSubdiv2DPoint* dst_pt;
	Vec2f pos_org, pos_dst;
	CvPoint2D32f org, dst;
	
	CvSeqReader  reader;
	cvStartReadSeq( (CvSeq*)(subdiv->edges), &reader, 0 );
	
	for( i = 0; i < total; i++ ){
		CvQuadEdge2D* edge = (CvQuadEdge2D*)(reader.ptr);
		
		if( CV_IS_SET_ELEM( edge )){
			org_pt = cvSubdiv2DEdgeOrg((CvSubdiv2DEdge)edge);
			dst_pt = cvSubdiv2DEdgeDst((CvSubdiv2DEdge)edge);
			if( org_pt && dst_pt ){
				org = org_pt->pt;
				dst = dst_pt->pt;
				pos_org = Vec2f(org.x, org.y);
				pos_dst = Vec2f(dst.x, dst.y);
				org_id = org_pt->flags;
				dst_id = dst_pt->flags;
				if(points[org_id] == pos_org && points[dst_id] == pos_dst){
					temp_lines.push_back(LinePos(points[org_id], points[dst_id]));
				}
			}
		}
		
		CV_NEXT_SEQ_ELEM( elem_size, reader );
	}
	
	return temp_lines;
}

CINDER_APP_BASIC( DelaunayLineApp, RendererGl )
